from .calc_iou import box_iou


__all__ = ['CalcAP']


class CalcAP:
    def __init__(self, labels, preds, iou_thr=0.5, conf_thr=0.0):
        self._labels = labels
        self._preds = preds
        self._iou_thr = iou_thr
        self._conf_thr = conf_thr
        self._precisions = dict()
        self._recalls = dict()

        pr_data = self._calc_pr()
        self._precisions['origin'] = pr_data['precisions']
        self._recalls['origin'] = pr_data['recalls']
        self._TP = pr_data['TP']
        self._FP = pr_data['FP']
        self._FN = len(self._labels) - self._TP
        self._tp_avg_iou = pr_data['tp_avg_iou']
        self._precision = self._TP / (self._TP + self._FP) if (self._TP + self._FP) else 0.0
        self._recall = self._TP / len(self._labels)
        if (self._precision + self._recall) == 0:
            self._f1_score = 0.0
        else:
            self._f1_score = 2 * (self._precision * self._recall) / (self._precision + self._recall)
        
        voc_ap_data = self._calc_voc_ap()
        self._AP = voc_ap_data['AP']
        self._precisions['interpolated'] = voc_ap_data['interpolated_precisions']
        self._recalls['interpolated'] = voc_ap_data['interpolated_recalls']

    def get(self):
        ret = {
            'n_labels': len(self._labels),
            'TP': self._TP,
            'FP': self._FP,
            'FN': self._FN,
            'tp_avg_iou': self._tp_avg_iou,
            'precision': self._precision,
            'recall': self._recall,
            'f1_score': self._f1_score,
            'precisions': self._precisions,
            'recalls': self._recalls,
            'AP': self._AP,
        }
        return ret
        
    def _calc_ious_confs_useds(self):
        labels_mapped = self._labels.map_by_img_id_with_used()
        ious, confs, useds = list(), list(), list()
        for pred in self._preds:
            *pts, conf, img_id = pred

            img_labels = labels_mapped.get(img_id, [])
            img_ious = list()
            for img_label in img_labels:
                label_pts, used = img_label.get('points'), img_label.get('used')
                iou = box_iou(pts, label_pts)
                img_ious.append(iou)

            if img_ious:
                max_iou = max(img_ious)
                mapped_label_idx = img_ious.index(max_iou)
                used = img_labels[mapped_label_idx].get('used')
                if max_iou >= self._iou_thr and not used:
                    img_labels[mapped_label_idx]['used'] = True
            else:
                max_iou, used = 0.0, False

            ious.append(max_iou)
            confs.append(conf)
            useds.append(used)
        return ious, confs, useds

    def _calc_pr(self):
        TP, FP = 0, 0
        precisions, recalls = [0.0], [0.0]
        tp_ious_sum = 0

        ious, confs, useds = self._calc_ious_confs_useds()
        for iou, conf, used in zip(ious, confs, useds):
            if iou >= self._iou_thr and conf >= self._conf_thr and not used:
                TP += 1
                tp_ious_sum += iou
            else:
                FP += 1
                
            precision = TP / (TP + FP)
            recall = TP / len(self._labels)
            precisions.append(precision)
            recalls.append(recall)
            
        precisions.append(0.0)
        recalls.append(1.0)
        tp_avg_iou = (tp_ious_sum / TP) if TP else 0.0

        ret = {
            'precisions': precisions,
            'recalls': recalls,
            'TP': TP,
            'FP': FP,
            'tp_avg_iou': tp_avg_iou,
        }
        return ret

    def _calc_voc_ap(self):
        precisions = self._precisions['origin'].copy()
        recalls = self._recalls['origin'].copy()
        
        # Interpolation
        for i in range(len(precisions)-2, -1, -1):
            precisions[i] = max(precisions[i], precisions[i+1])

        recalls_changed_idx = list()
        for idx in range(1, len(recalls)):
            if recalls[idx] != recalls[idx-1]:
                recalls_changed_idx.append(idx)

        ap = 0.0
        for idx in recalls_changed_idx:
            recalls_diff = recalls[idx] - recalls[idx-1]
            ap += (recalls_diff * precisions[idx])

        ret = {
            'AP': ap,
            'interpolated_precisions': precisions,
            'interpolated_recalls': recalls,
        }
        return ret



from collections import defaultdict


__all__ = ['classify_labels', 'classify_preds']


def classify_labels(labels):
    """
    Args:
        labels (list): list of [left, top, right, bottom, class_name, image_id]

    Returns:
        defaultdict: {key: class_name (str), value: ClassLabel Object}
    """
    labels_classified = defaultdict(_ClassLabel)
    for label in labels:
        *pts, cls_name, img_id = label
        labels_classified[cls_name].labels.append([*pts, img_id])
    return labels_classified


def classify_preds(preds):
    """
    Args:
        preds (list): list of [left, top, right, bottom, confidence, class_name, image_id]

    Returns:
        dict: {key: class_name (str), value: list of [left, top, right, bottom, confidence, image_id]}
    """
    preds_classified = defaultdict(list)
    for pred in preds:
        *pts, confidence, cls_name, img_id = pred
        preds_classified[cls_name].append([*pts, confidence, img_id])

    # Descending Ordering with Confidence
    for cls_name in preds_classified:
        preds_classified[cls_name].sort(key=lambda x : x[-2], reverse=True)
    return preds_classified


class _ClassLabel:
    def __init__(self):
        self.labels = list()
    
    def __len__(self):
        return len(self.labels)

    def map_by_img_id_with_used(self):
        labels_mapped = defaultdict(list)
        for label in self.labels:
            *pts, img_id = label
            labels_mapped[img_id].append({'points': pts, 'used': False})
        return labels_mapped



import os
from setuptools import setup


ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
README_PATH = os.path.join(ROOT_DIR, 'README.md')

setup(
    name='calc4ap',
    author='devbruce',
    author_email='bruce93k@gmail.com',
    description='Easy AP Calculator with Python',
    long_description=open(README_PATH, encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    license='MIT',
    version='1.0.1',
    url='https://github.com/DevBruce/calc4ap',
    packages=['calc4ap'],
    package_data={
        'calc4ap': ['libs/*.py', 'utils/*.py'],
    },
    keywords=[
        'ap',
        'map',
        'object_detection',
        'pascal_voc',
        'coco',
        'deep_learning',
        'evaluation',
    ],
    python_requires='>=3.6.5',
    install_requires=[],
)



__all__ = ['box_iou']


def box_iou(box1_pts, box2_pts):
    """Calculate IoU of box1 and box2

    Args:
        box1_pts (list): [left, top, right, bottom] (xyrb)
        box2_pts (list): [left, top, right, bottom] (xyrb)

    Returns:
        float: IoU of box1 and box2
    """
    box_intersection_pts = [
        max(box1_pts[0], box2_pts[0]),
        max(box1_pts[1], box2_pts[1]),
        min(box1_pts[2], box2_pts[2]),
        min(box1_pts[3], box2_pts[3]),
        ]
    intersection_width = box_intersection_pts[2] - box_intersection_pts[0] + 1
    intersection_height = box_intersection_pts[3] - box_intersection_pts[1] + 1
    intersection_area = intersection_width * intersection_height
    
    if intersection_width > 0 and intersection_height > 0:
        box1_width = box1_pts[2] - box1_pts[0] + 1
        box1_height = box1_pts[3] - box1_pts[1] + 1
        box2_width = box2_pts[2] - box2_pts[0] + 1
        box2_height = box2_pts[3] - box2_pts[1] + 1
        
        box1_area = box1_width * box1_height
        box2_area = box2_width * box2_height
        union_area = box1_area + box2_area - intersection_area
        iou = intersection_area / union_area
    else:
        iou = 0.0
    return iou



from .libs.classifiers import classify_labels, classify_preds
from .libs.calc_ap import CalcAP


__all__ = ['CalcVOCmAP']


class CalcVOCmAP:
    def __init__(self, labels, preds, iou_thr=0.5, conf_thr=0.0):
        self._labels = classify_labels(labels)
        self._preds = classify_preds(preds)
        self._n_classes = len(self._labels)
        self._iou_thr = iou_thr
        self._conf_thr = conf_thr
        self._APs = self._calc_APs()

        mAP_data = self._calc_mAP()
        self._mAP = mAP_data['mAP']
        self._w_mAP = mAP_data['w_mAP']
        self._total_TP = mAP_data['total_TP']
        self._total_FP = mAP_data['total_FP']
        self._total_FN = mAP_data['total_FN']
        self._m_tp_avg_iou = mAP_data['m_tp_avg_iou']
        self._m_precision = mAP_data['m_precision']
        self._m_recall = mAP_data['m_recall']
        self._m_f1_score = mAP_data['m_f1_score']

    def get(self):
        ret = {
            'total_TP': self._total_TP,
            'total_FP': self._total_FP,
            'total_FN': self._total_FN,
            'm_tp_avg_iou': self._m_tp_avg_iou,
            'm_precision': self._m_precision,
            'm_recall': self._m_recall,
            'm_f1_score': self._m_f1_score,
            'APs': self._APs,
            'mAP': self._mAP,
            'w_mAP': self._w_mAP,
        }
        return ret

    def get_summary(self):
        summary = dict()
        for cls_name in self._APs:
            summary[cls_name] = self._APs[cls_name]['AP']
        summary['mAP'] = self._mAP
        return summary

    def _calc_APs(self):
        APs = dict()
        for cls_name in self._labels:
            AP = CalcAP(
                labels=self._labels[cls_name],
                preds=self._preds[cls_name],
                iou_thr=self._iou_thr,
                conf_thr=self._conf_thr,
            ).get()
            APs[cls_name] = AP
        return APs

    def _calc_mAP(self):
        APs = 0.0
        TPs, FPs, FNs = 0, 0, 0
        sum_tp_avg_iou = 0.0
        sum_n_labels = 0
        for cls_name in self._labels:
            APs += self._APs[cls_name]['AP']
            TPs += self._APs[cls_name]['TP']
            FPs += self._APs[cls_name]['FP']
            FNs += self._APs[cls_name]['FN']
            sum_n_labels += self._APs[cls_name]['n_labels']
            sum_tp_avg_iou += self._APs[cls_name]['tp_avg_iou']

        mAP = APs / self._n_classes
        m_tp_avg_iou = sum_tp_avg_iou / self._n_classes
        m_precision = TPs / (TPs + FPs) if (TPs + FPs) else 0.0
        m_recall = TPs / sum_n_labels
        if (m_precision + m_recall) == 0:
            m_f1_score = 0.0
        else:
            m_f1_score = 2 * (m_precision * m_recall) / (m_precision + m_recall)
        
        # Weighted mAP
        w_APs = list()
        for cls_name in self._labels:
            w_AP = self._APs[cls_name]['AP'] * self._APs[cls_name]['n_labels']
            w_APs.append(w_AP)
        w_mAP = sum(w_APs) / sum_n_labels

        ret = {
            'mAP': mAP,
            'w_mAP': w_mAP,
            'total_TP': TPs,
            'total_FP': FPs,
            'total_FN': FNs,
            'm_tp_avg_iou': m_tp_avg_iou,
            'm_precision': m_precision,
            'm_recall': m_recall,
            'm_f1_score': m_f1_score,
        }
        return ret



import json


__all__ = ['get_labels_from_coco_ann']


def get_labels_from_coco_ann(coco_ann_path):
    coco_ann = json.load(open(coco_ann_path))
    cats = coco_ann['categories']
    anns = coco_ann['annotations']

    cls_map = dict()
    for cat in cats:
        cls_num = cat['id']
        cls_name = cat['name']
        cls_map[cls_num] = cls_name

    labels = list()
    for ann in anns:
        img_id = ann['image_id']
        cat_id = ann['category_id']
        cls_name = cls_map[cat_id]

        x, y, w, h = ann.get('bbox')
        left, top, right, bottom = x, y, x+w, y+h
        labels.append([left, top, right, bottom, cls_name, img_id])
    return labels
